function [err, timepoints, species_out, observables_out] = MAPK( timepoints, species_init, parameters, suppress_plot )
%MAPK Integrate reaction network and plot observables.
%   Integrates the reaction network corresponding to the BioNetGen model
%   'MAPK' and then (optionally) plots the observable trajectories,
%   or species trajectories if no observables are defined. Trajectories are
%   generated using either default or user-defined parameters and initial
%   species values. Integration is performed by the MATLAB stiff solver
%   'ode15s'. MAPK returns an error value, a vector of timepoints,
%   species trajectories, and observable trajectories.
%   
%   [err, timepoints, species_out, observables_out]
%        = MAPK( timepoints, species_init, parameters, suppress_plot )
%
%   INPUTS:
%   -------
%   species_init    : row vector of 34 initial species populations.
%   timepoints      : column vector of time points returned by integrator.
%   parameters      : row vector of 24 model parameters.
%   suppress_plot   : 0 if a plot is desired (default), 1 if plot is suppressed.
%
%   Note: to specify default value for an input argument, pass the empty array.
%
%   OUTPUTS:
%   --------
%   err             : 0 if the integrator exits without error, non-zero otherwise.
%   timepoints      : a row vector of timepoints returned by the integrator.
%   species_out     : array of species population trajectories
%                        (columns correspond to species, rows correspond to time).
%   observables_out : array of observable trajectories
%                        (columns correspond to observables, rows correspond to time).
%
%   QUESTIONS about the BNG Mfile generator?  Email justinshogg@gmail.com



%% Process input arguments

% define any missing arguments
if ( nargin < 1 )
    timepoints = [];
end

if ( nargin < 2 )
    species_init = [];
end

if ( nargin < 3 )
    parameters = [];
end

if ( nargin < 4 )
    suppress_plot = 0;
end


% initialize outputs (to avoid error msgs if script terminates early
err = 0;
species_out     = [];
observables_out = [];


% setup default parameters, if necessary
if ( isempty(parameters) )
   parameters = [ 10, 1.0, 1.0, 1.0, 1.00, 0, 1e-2, 1e-2, 1e+2, 1, 1, 1e-2, 1e+2, 1e-2, 1e-2, 1e-2, 3e-4, 3e-4, 1e+2, 3e-4, 3e-9, 2e-3, 1e-9, 1e-3 ];
end
% check that parameters has proper dimensions
if (  size(parameters,1) ~= 1  ||  size(parameters,2) ~= 24  )
    fprintf( 1, 'Error: size of parameter argument is invalid! Correct size = [1 24].\n' );
    err = 1;
    return;
end

% setup default initial values, if necessary
if ( isempty(species_init) )
   species_init = initialize_species( parameters );
end
% check that species_init has proper dimensions
if (  size(species_init,1) ~= 1  ||  size(species_init,2) ~= 34  )
    fprintf( 1, 'Error: size of species_init argument is invalid! Correct size = [1 34].\n' );
    err = 1;
    return;
end

% setup default timepoints, if necessary
if ( isempty(timepoints) )
   timepoints = linspace(0,24*3600,24*60+1)';
end
% check that timepoints has proper dimensions
if (  size(timepoints,1) < 2  ||  size(timepoints,2) ~= 1  )
    fprintf( 1, 'Error: size of timepoints argument is invalid! Correct size = [t 1], t>1.\n' );
    err = 1;
    return;
end

% setup default suppress_plot, if necessary
if ( isempty(suppress_plot) )
   suppress_plot = 0;
end
% check that suppress_plot has proper dimensions
if ( size(suppress_plot,1) ~= 1  ||  size(suppress_plot,2) ~= 1 )
    fprintf( 1, 'Error: suppress_plots argument should be a scalar!\n' );
    err = 1;
    return;
end

% define parameter labels (this is for the user's reference!)
param_labels = { 'EGF', 'ERKpp_SOS1_FB', 'ERKpp_MEK_FB', 'ERKpp_RAF1_FB', 'BIMOL', 'RAS_t0_active', 'd1', 'u1a', 'u1b', 'u2a', 'u2b', 'u3', 'k3', 'd2', 'q1', 'q2', 'q3', 'q4', 'q5', 'q6', 'a0_ekarev', 'd0_ekarev', 'a0_erktr', 'd0_erktr' };



%% Integrate Network Model
 
% calculate expressions
[expressions] = calc_expressions( parameters );

% set ODE integrator options
opts = odeset( 'RelTol',   1e-8,   ...
               'AbsTol',   0.0001,   ...
               'Stats',    'off',  ...
               'BDF',      'off',    ...
               'MaxOrder', 5   );


% define derivative function
rhs_fcn = @(t,y)( calc_species_deriv( t, y, expressions ) );

% simulate model system (stiff integrator)
try 
    [~, species_out] = ode15s( rhs_fcn, timepoints, species_init', opts );
    if(length(timepoints) ~= size(species_out,1))
        exception = MException('ODE15sError:MissingOutput','Not all timepoints output\n');
        throw(exception);
    end
catch
    err = 1;
    fprintf( 1, 'Error: some problem encountered while integrating ODE network!\n' );
    return;
end

% calculate observables
observables_out = zeros( length(timepoints), 6 );
for t = 1 : length(timepoints)
    observables_out(t,:) = calc_observables( species_out(t,:), expressions );
end


%% Plot Output, if desired

if ( ~suppress_plot )
    
    % define plot labels
    observable_labels = { 'EGFR_act', 'EGFR_act_SOS_act_cplx', 'RAS_act', 'RAF_act', 'MEK_act', 'ERK_act' };

    % construct figure
    plot(timepoints,observables_out);
    title('MAPK observables','fontSize',14,'Interpreter','none');
    axis([0 timepoints(end) 0 inf]);
    legend(observable_labels,'fontSize',10,'Interpreter','none');
    xlabel('time','fontSize',12,'Interpreter','none');
    ylabel('number or concentration','fontSize',12,'Interpreter','none');

end


%~~~~~~~~~~~~~~~~~~~~~%
% END of main script! %
%~~~~~~~~~~~~~~~~~~~~~%


% initialize species function
function [species_init] = initialize_species( params )

    species_init = zeros(1,34);
    species_init(1) = 3e5*params(5);
    species_init(2) = (6e4*params(5))*(1-params(6));
    species_init(3) = (6e4*params(5))*params(6);
    species_init(4) = 1e5*params(5);
    species_init(5) = 6e3*params(5);
    species_init(6) = 5e5*params(5);
    species_init(7) = 2e5*params(5);
    species_init(8) = 3e6*params(5);
    species_init(9) = 1e6*params(5);
    species_init(10) = 1e6*params(5);
    species_init(11) = 0;
    species_init(12) = 0;
    species_init(13) = 0;
    species_init(14) = 0;
    species_init(15) = 0;
    species_init(16) = 0;
    species_init(17) = 0;
    species_init(18) = 0;
    species_init(19) = 0;
    species_init(20) = 0;
    species_init(21) = 0;
    species_init(22) = 0;
    species_init(23) = 0;
    species_init(24) = 0;
    species_init(25) = 0;
    species_init(26) = 0;
    species_init(27) = 0;
    species_init(28) = 0;
    species_init(29) = 0;
    species_init(30) = 0;
    species_init(31) = 0;
    species_init(32) = 0;
    species_init(33) = 0;
    species_init(34) = 0;

end


% user-defined functions



% Calculate expressions
function [ expressions ] = calc_expressions ( parameters )

    expressions = zeros(1,59);
    expressions(1) = parameters(1);
    expressions(2) = parameters(2);
    expressions(3) = parameters(3);
    expressions(4) = parameters(4);
    expressions(5) = parameters(5);
    expressions(6) = parameters(6);
    expressions(7) = (3e5*expressions(5));
    expressions(8) = (6e4*expressions(5));
    expressions(9) = (1e5*expressions(5));
    expressions(10) = (6e3*expressions(5));
    expressions(11) = (5e5*expressions(5));
    expressions(12) = (2e5*expressions(5));
    expressions(13) = (3e6*expressions(5));
    expressions(14) = (1e6*expressions(5));
    expressions(15) = (1e6*expressions(5));
    expressions(16) = (5e-5*expressions(1));
    expressions(17) = parameters(7);
    expressions(18) = (1e-5/expressions(5));
    expressions(19) = parameters(8);
    expressions(20) = parameters(9);
    expressions(21) = (1e-6/expressions(5));
    expressions(22) = parameters(10);
    expressions(23) = (1e-7/expressions(5));
    expressions(24) = parameters(11);
    expressions(25) = (1e-4/expressions(5));
    expressions(26) = (1e-5/expressions(5));
    expressions(27) = (1e-5/expressions(5));
    expressions(28) = parameters(12);
    expressions(29) = parameters(13);
    expressions(30) = (1e-7/expressions(5));
    expressions(31) = parameters(14);
    expressions(32) = (1e-7/expressions(5));
    expressions(33) = parameters(15);
    expressions(34) = (3e-6/expressions(5));
    expressions(35) = parameters(16);
    expressions(36) = ((3e-9/expressions(5))*expressions(2));
    expressions(37) = parameters(17);
    expressions(38) = ((6e-10/expressions(5))*expressions(3));
    expressions(39) = parameters(18);
    expressions(40) = parameters(19);
    expressions(41) = ((6e-10/expressions(5))*expressions(4));
    expressions(42) = parameters(20);
    expressions(43) = parameters(21);
    expressions(44) = parameters(22);
    expressions(45) = parameters(23);
    expressions(46) = parameters(24);
    expressions(47) = (expressions(8)*(1-expressions(6)));
    expressions(48) = (expressions(8)*expressions(6));
    expressions(49) = (expressions(32)*2);
    expressions(50) = (expressions(33)*2);
    expressions(51) = (expressions(34)*2);
    expressions(52) = (expressions(35)*2);
    expressions(53) = (expressions(36)*4);
    expressions(54) = (expressions(36)*3);
    expressions(55) = (expressions(36)*2);
    expressions(56) = (expressions(37)*2);
    expressions(57) = (expressions(37)*3);
    expressions(58) = (expressions(37)*4);
    expressions(59) = (expressions(40)*2);
   
end



% Calculate observables
function [ observables ] = calc_observables ( species, expressions )

    observables = zeros(1,6);
    observables(1) = species(11) +species(14) +species(18) +species(19);
    observables(2) = species(14) +species(18) +species(19);
    observables(3) = species(3) +species(18) +species(21);
    observables(4) = species(13);
    observables(5) = species(20) +species(28);
    observables(6) = species(24);

end


% Calculate ratelaws
function [ ratelaws ] = calc_ratelaws ( species, expressions, observables )

    ratelaws = zeros(1,6);
    ratelaws(1) = (5e-5*expressions(1))*species(1);
    ratelaws(2) = (1e-5/expressions(5))*species(5)*species(3);
    ratelaws(3) = (1e-7/expressions(5))*species(3)*species(6);
    ratelaws(4) = expressions(17)*species(11);
    ratelaws(5) = (1e-5/expressions(5))*species(11)*species(4);
    ratelaws(6) = expressions(29)*species(12);
    ratelaws(7) = expressions(31)*species(13);
    ratelaws(8) = (expressions(32)*2)*species(13)*species(7);
    ratelaws(9) = expressions(17)*species(14);
    ratelaws(10) = expressions(19)*species(14);
    ratelaws(11) = (1e-6/expressions(5))*species(14)*species(3);
    ratelaws(12) = (1e-7/expressions(5))*species(14)*species(2);
    ratelaws(13) = expressions(28)*species(15);
    ratelaws(14) = (1e-7/expressions(5))*species(13)*species(16);
    ratelaws(15) = expressions(33)*species(16);
    ratelaws(16) = (5e-5*expressions(1))*species(17);
    ratelaws(17) = expressions(17)*species(18);
    ratelaws(18) = expressions(17)*species(19);
    ratelaws(19) = expressions(20)*species(17);
    ratelaws(20) = (1e-6/expressions(5))*species(17)*species(3);
    ratelaws(21) = expressions(22)*species(18);
    ratelaws(22) = (1e-7/expressions(5))*species(17)*species(2);
    ratelaws(23) = expressions(24)*species(19);
    ratelaws(24) = (1e-4/expressions(5))*species(18)*species(2);
    ratelaws(25) = (1e-5/expressions(5))*species(19)*species(2);
    ratelaws(26) = (expressions(33)*2)*species(20);
    ratelaws(27) = (expressions(34)*2)*species(20)*species(8);
    ratelaws(28) = (5e-5*expressions(1))*species(21);
    ratelaws(29) = (5e-5*expressions(1))*species(22);
    ratelaws(30) = expressions(22)*species(21);
    ratelaws(31) = expressions(24)*species(22);
    ratelaws(32) = (1e-4/expressions(5))*species(21)*species(2);
    ratelaws(33) = (1e-5/expressions(5))*species(22)*species(2);
    ratelaws(34) = (3e-6/expressions(5))*species(20)*species(23);
    ratelaws(35) = expressions(35)*species(23);
    ratelaws(36) = (expressions(35)*2)*species(24);
    ratelaws(37) = (expressions(36)*4)*species(24)*species(4);
    ratelaws(38) = ((6e-10/expressions(5))*expressions(3))*species(24)*species(7);
    ratelaws(39) = ((6e-10/expressions(5))*expressions(3))*species(24)*species(16);
    ratelaws(40) = ((6e-10/expressions(5))*expressions(3))*species(24)*species(20);
    ratelaws(41) = ((6e-10/expressions(5))*expressions(4))*species(24)*species(13);
    ratelaws(42) = ((6e-10/expressions(5))*expressions(4))*species(24)*species(6);
    ratelaws(43) = expressions(43)*species(24)*species(9);
    ratelaws(44) = expressions(45)*species(24)*species(10);
    ratelaws(45) = (expressions(32)*2)*species(13)*species(26);
    ratelaws(46) = (1e-7/expressions(5))*species(13)*species(27);
    ratelaws(47) = (expressions(33)*2)*species(28);
    ratelaws(48) = expressions(33)*species(27);
    ratelaws(49) = (expressions(34)*2)*species(28)*species(8);
    ratelaws(50) = (3e-6/expressions(5))*species(28)*species(23);
    ratelaws(51) = (expressions(36)*3)*species(24)*species(25);
    ratelaws(52) = expressions(37)*species(25);
    ratelaws(53) = expressions(39)*species(26);
    ratelaws(54) = expressions(39)*species(27);
    ratelaws(55) = expressions(39)*species(28);
    ratelaws(56) = (expressions(40)*2)*species(28);
    ratelaws(57) = expressions(40)*species(27);
    ratelaws(58) = expressions(42)*species(29);
    ratelaws(59) = expressions(44)*species(30);
    ratelaws(60) = expressions(46)*species(31);
    ratelaws(61) = (expressions(36)*2)*species(24)*species(32);
    ratelaws(62) = (expressions(37)*2)*species(32);
    ratelaws(63) = ((3e-9/expressions(5))*expressions(2))*species(24)*species(33);
    ratelaws(64) = (expressions(37)*3)*species(33);
    ratelaws(65) = (expressions(37)*4)*species(34);

end

% Calculate species derivatives
function [ Dspecies ] = calc_species_deriv ( time, species, expressions )
    
    % initialize derivative vector
    Dspecies = zeros(34,1);
    
    % update observables
    [ observables ] = calc_observables( species, expressions );
    
    % update ratelaws
    [ ratelaws ] = calc_ratelaws( species, expressions, observables );
                        
    % calculate derivatives
    Dspecies(1) = -ratelaws(1) +ratelaws(4) +ratelaws(19);
    Dspecies(2) = -ratelaws(12) +ratelaws(13) -ratelaws(22) +ratelaws(23) -ratelaws(24) -ratelaws(25) +ratelaws(31) -ratelaws(32) -ratelaws(33);
    Dspecies(3) = -ratelaws(2) -ratelaws(11) -ratelaws(20) +ratelaws(21) +ratelaws(24) +ratelaws(25) +ratelaws(30) +ratelaws(32) +ratelaws(33);
    Dspecies(4) = -ratelaws(5) +ratelaws(10) +ratelaws(19) -ratelaws(37) +ratelaws(52);
    Dspecies(5) = -ratelaws(2) +ratelaws(13);
    Dspecies(6) = -ratelaws(3) +ratelaws(7) -ratelaws(42) +ratelaws(58);
    Dspecies(7) = -ratelaws(8) +ratelaws(15) -ratelaws(38) +ratelaws(53);
    Dspecies(8) = -ratelaws(27) +ratelaws(35) -ratelaws(49);
    Dspecies(9) = -ratelaws(43) +ratelaws(59);
    Dspecies(10) = -ratelaws(44) +ratelaws(60);
    Dspecies(11) = ratelaws(1) -ratelaws(4) -ratelaws(5) +ratelaws(10);
    Dspecies(12) = ratelaws(2) -ratelaws(6);
    Dspecies(13) = ratelaws(3) -ratelaws(7) -ratelaws(41);
    Dspecies(14) = ratelaws(5) -ratelaws(9) -ratelaws(10) -ratelaws(11) -ratelaws(12) +ratelaws(16) +ratelaws(21) +ratelaws(23);
    Dspecies(15) = ratelaws(6) -ratelaws(13);
    Dspecies(16) = ratelaws(8) -ratelaws(14) -ratelaws(15) +ratelaws(26) -ratelaws(39) +ratelaws(54);
    Dspecies(17) = ratelaws(9) -ratelaws(16) -ratelaws(19) -ratelaws(20) -ratelaws(22) +ratelaws(30) +ratelaws(31);
    Dspecies(18) = ratelaws(11) -ratelaws(17) -ratelaws(21) +ratelaws(28);
    Dspecies(19) = ratelaws(12) -ratelaws(18) -ratelaws(23) +ratelaws(29);
    Dspecies(20) = ratelaws(14) -ratelaws(26) -ratelaws(40) +ratelaws(55);
    Dspecies(21) = ratelaws(17) +ratelaws(20) -ratelaws(28) -ratelaws(30);
    Dspecies(22) = ratelaws(18) +ratelaws(22) -ratelaws(29) -ratelaws(31);
    Dspecies(23) = ratelaws(27) -ratelaws(34) -ratelaws(35) +ratelaws(36) +ratelaws(49) -ratelaws(50);
    Dspecies(24) = ratelaws(34) -ratelaws(36) +ratelaws(50);
    Dspecies(25) = ratelaws(37) -ratelaws(51) -ratelaws(52) +ratelaws(62);
    Dspecies(26) = ratelaws(38) -ratelaws(45) +ratelaws(48) -ratelaws(53) +ratelaws(57);
    Dspecies(27) = ratelaws(39) +ratelaws(45) -ratelaws(46) +ratelaws(47) -ratelaws(48) -ratelaws(54) +ratelaws(56) -ratelaws(57);
    Dspecies(28) = ratelaws(40) +ratelaws(46) -ratelaws(47) -ratelaws(55) -ratelaws(56);
    Dspecies(29) = ratelaws(41) +ratelaws(42) -ratelaws(58);
    Dspecies(30) = ratelaws(43) -ratelaws(59);
    Dspecies(31) = ratelaws(44) -ratelaws(60);
    Dspecies(32) = ratelaws(51) -ratelaws(61) -ratelaws(62) +ratelaws(64);
    Dspecies(33) = ratelaws(61) -ratelaws(63) -ratelaws(64) +ratelaws(65);
    Dspecies(34) = ratelaws(63) -ratelaws(65);

end


end
